"""Tests for helper classes."""

import httpx
import pytest
from pytest_httpx import HTTPXMock

from notifications_android_tv import ImageSource
from notifications_android_tv import ConnectError, InvalidImage


@pytest.mark.asyncio
async def test_image_source() -> None:
    """Test constructing ImageSource."""
    # test provding wrong authentication type
    with pytest.raises(ValueError) as err:
        ImageSource.from_url(url="http://example.com/image.png", auth="something")
        assert err == "authentication must be 'basic' or 'digest'"

    # test missing password
    with pytest.raises(ValueError) as err:
        ImageSource.from_url(
            url="http://example.com/image.png", auth="basic", username="user"
        )
        assert err == "username and password must be specified"

    # test missing username
    with pytest.raises(ValueError) as err:
        ImageSource.from_url(
            "http://example.com/image.png", auth="basic", password="pass"
        )
        assert err == "username and password must be specified"

    # test url with basic auth
    image_source_dict = {
        "url": "http://example.com/image.png",
        "auth": "basic",
        "username": "user",
        "password": "pass",
    }
    image_source = ImageSource.from_url(**image_source_dict)
    assert image_source.url == "http://example.com/image.png"
    assert isinstance(image_source.auth, httpx.BasicAuth)

    # test url with digest auth
    image_source_dict = {
        "url": "http://example.com/image.png",
        "auth": "digest",
        "username": "user",
        "password": "pass",
    }
    image_source = ImageSource.from_url(**image_source_dict)
    assert image_source.url == "http://example.com/image.png"
    assert isinstance(image_source.auth, httpx.DigestAuth)


@pytest.mark.asyncio
async def test_get_image_fails(httpx_mock: HTTPXMock) -> None:
    """Test getting an image from source fails."""
    image = ImageSource.from_url(url="http://example.com")

    # test timeout fetching image
    with pytest.raises(ConnectError):
        httpx_mock.add_exception(httpx.TimeoutException(""))
        await image.async_get_image()

    # test image url doesn't return 200
    httpx_mock.add_response(status_code=400)
    with pytest.raises(InvalidImage):
        await image.async_get_image()

    # test returned content is not an image type
    httpx_mock.add_response(headers={"content-type": "text/html"})
    with pytest.raises(InvalidImage):
        await image.async_get_image()

    # test getting image non existing file fails
    image2 = ImageSource.from_path("image_file.jpg")
    with pytest.raises(InvalidImage):
        await image2.async_get_image()



"""Helper methods for the library."""

from __future__ import annotations

import base64
from dataclasses import dataclass
from typing import Any

import httpx

from .const import (
    DEFAULT_BKGCOLOR,
    DEFAULT_DURATION,
    DEFAULT_FONTSIZE,
    DEFAULT_ICON,
    DEFAULT_POSITION,
    DEFAULT_TRANSPARENCY,
    BkgColor,
    FontSize,
    Position,
    Transparency,
)
from .exceptions import ConnectError, InvalidImage, InvalidImageData


@dataclass
class ImageSource:
    """Image source from url or local path."""

    path: str | None = None
    url: str | None = None
    auth: httpx.Auth | None = None

    @classmethod
    def from_path(cls, path: str) -> ImageSource:
        """Initiate image source class."""

        return cls(path=path)

    @classmethod
    def from_url(
        cls,
        url: str,
        username: str | None = None,
        password: str | None = None,
        auth: str | None = None,
    ) -> ImageSource:
        """Initiate image source class."""
        _cls = cls(url=url)
        if auth:
            if auth not in ["basic", "digest"]:
                raise ValueError(f"Invalid auth '{auth}', must be 'basic' or 'digest'")
            if username is None or password is None:
                raise ValueError("username and password must be specified")
            if auth == "basic":
                _cls.auth = httpx.BasicAuth(username, password)
            else:
                _cls.auth = httpx.DigestAuth(username, password)

        return _cls

    async def async_get_image(self) -> bytes:
        """Load file from path or url."""
        if self.path is not None:
            try:
                with open(self.path, "rb") as file:
                    return file.read()
            except FileNotFoundError as err:
                raise InvalidImage(err) from err
        if self.url is not None:
            try:
                async with httpx.AsyncClient(verify=False) as client:
                    response = await client.get(self.url, auth=self.auth, timeout=30)

            except (httpx.ConnectError, httpx.TimeoutException) as err:
                raise ConnectError(
                    f"Error fetching image from {self.url}: {err}"
                ) from err
            if response.status_code != httpx.codes.OK:
                raise InvalidImage(f"Error fetching image from {self.url}: {response}")
            if "image" not in response.headers["content-type"]:
                raise InvalidImage(
                    f"Response content type is not an image: {response.headers['content-type']}"
                )
            return response.content
        raise ValueError("Either path or url must be specified")


@dataclass
class NotificationParams:
    """Notification parameters.

    :param duration: (Optional) Display the notification for the specified period.
        Default duration is 5 seconds.
    :param position: (Optional) Specify notification position from class Position.
        Default is `Positions.BOTTOM_RIGHT`.
    :param fontsize: (Optional) Specify text font size from class FontSize.
        Default is `FontSizes.MEDIUM`.
    :param color: (Optional) Specify background color from class BkgColor.
        Default is `BkgColors.GREY`.
    :param transparency: (Optional) Specify the background transparency of the notification
        from class `Transparency`. Default is 0%.
    :param interrupt: (Optional) Setting it to true makes the notification interactive
        and can be dismissed or selected to display more details. Default is False
    :param icon: (Optional) Attach icon to notification. Construct using ImageSource.
    :param image_file: (Optional) Attach image to notification. Construct using ImageSource.
    """

    duration: int = DEFAULT_DURATION
    position: int = DEFAULT_POSITION
    fontsize: FontSize = DEFAULT_FONTSIZE
    transparency: int = DEFAULT_TRANSPARENCY
    color: BkgColor = DEFAULT_BKGCOLOR
    interrupt: bool = False
    icon: ImageSource | None = None
    image: ImageSource | None = None

    @property
    def data_params(self) -> dict[str, Any]:
        """Return notification parameters as a dict."""
        return {
            "duration": self.duration,
            "position": self.position,
            "fontsize": self.fontsize.value,
            "transparency": self.transparency,
            "color": self.color.value,
            "interrupt": self.interrupt,
        }

    async def get_images(self) -> dict[str, Any]:
        """Return notification icon and image."""
        icon_bytes = base64.b64decode(DEFAULT_ICON)
        if self.icon is not None:
            icon_bytes = await self.icon.async_get_image()
        files = {
            "filename": (
                "image",
                icon_bytes,
                "application/octet-stream",
                {"Expires": "0"},
            )
        }
        if self.image is not None:
            image_bytes = await self.image.async_get_image()
            files["filename2"] = (
                "image",
                image_bytes,
                "application/octet-stream",
                {"Expires": "0"},
            )
        return files

    @classmethod
    def from_dict(cls, kwargs: dict[str, Any]) -> NotificationParams:
        """Initiate notification parameters class."""
        _params: dict[str, Any] = {}
        if duration := kwargs.get("duration"):
            try:
                _params["duration"] = int(duration)
            except ValueError:
                _params["duration"] = DEFAULT_DURATION

        if position := kwargs.get("position"):
            _params["position"] = Position.from_string(position)

        if (font_size := kwargs.get("fontsize")) and hasattr(
            FontSize, font_size.upper()
        ):
            _params["fontsize"] = getattr(FontSize, font_size.upper())

        if transparency := kwargs.get("transparency"):
            _params["transparency"] = Transparency.from_percentage(transparency)

        if (color := kwargs.get("color")) and hasattr(BkgColor, color.upper()):
            _params["color"] = getattr(BkgColor, color.upper())

        if interrupt := kwargs.get("interrupt"):
            _params["interrupt"] = bool(interrupt)

        if icon := kwargs.get("icon"):
            _params["icon"] = create_image_source("icon", icon)
        if image := kwargs.get("image"):
            _params["image"] = create_image_source("image", image)

        return NotificationParams(**_params)


def create_image_source(key: str, data: dict[str, Any]) -> ImageSource:
    """create image source class."""
    if isinstance(data, str):
        return (
            ImageSource.from_url(data)
            if data.startswith("http")
            else ImageSource.from_path(data)
        )
    elif isinstance(data, dict) and "path" in data:
        return ImageSource.from_path(data["path"])
    elif (
        isinstance(data, dict) and (url := data.get("url")) and (url.startswith("http"))
    ):
        try:
            return ImageSource.from_url(**data)
        except ValueError as err:
            raise InvalidImageData(f"Invalid '{key}' data: {str(err)}") from err
    else:
        raise InvalidImageData(f"Invalid '{key}' data")



"""Library for sending notifications to Android/Fire TVs."""

from .notifier import Notifications
from .helpers import ImageSource, NotificationParams
from .exceptions import (
    ConnectError,
    InvalidImage,
    InvalidImageData,
    InvalidResponse,
    NotificationException,
)

__all__ = [
    "Notifications",
    "ImageSource",
    "NotificationParams",
    "ConnectError",
    "InvalidImage",
    "InvalidImageData",
    "InvalidResponse",
    "NotificationException",
]



"""Constants for the library."""

from enum import Enum, IntEnum
from typing import Final


class BkgColor(Enum):
    """Background color options."""

    GREY = "#607d8b"
    BLACK = "#000000"
    INDIGO = "#303F9F"
    GREEN = "#4CAF50"
    RED = "#F44336"
    CYAN = "#00BCD4"
    TEAL = "#009688"
    AMBER = "#FFC107"
    PINK = "#E91E63"


class FontSize(IntEnum):
    """Supported font sizes for notification text."""

    SMALL = 1
    MEDIUM = 0
    LARGE = 2
    MAX = 3


class Position:
    """Position of the notification.

    Supported values:
      - 0: Bottom right
      - 1: Bottom left
      - 2: Top right
      - 3: Top left
      - 4: Center
    """

    @classmethod
    def from_string(cls, position: str) -> int:
        """Convert position to int."""
        _mapping = {
            "bottom-right": 0,
            "bottom-left": 1,
            "top-right": 2,
            "top-left": 3,
            "center": 4,
        }
        return _mapping.get(position, 0)


class Transparency:
    """Transparency for the notification overlay.

    Supported values:
      - 1: 0%
      - 2; 25%
      - 3: 50%
      - 4: 75%
      - 5: 100%
    """

    @classmethod
    def from_percentage(cls, percentage: str) -> int:
        """Convert percentage to int."""
        _mapping = {
            "0%": 1,
            "25%": 2,
            "50%": 3,
            "75%": 4,
            "100%": 5,
        }
        return _mapping.get(percentage, 1)


DEFAULT_TITLE: Final = "Notification"
DEFAULT_DURATION: Final = 5
DEFAULT_POSITION: Final = 0
DEFAULT_BKGCOLOR: Final = BkgColor.GREY
DEFAULT_FONTSIZE: Final = FontSize.MEDIUM
DEFAULT_TRANSPARENCY: Final = 1
DEFAULT_ICON: Final = (
    "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR4nGP6zwAAAgcBApo"
    "cMXEAAAAASUVORK5CYII="
)



"""Tests for Notifications."""

import httpx
import pytest
from pytest_httpx import HTTPXMock

from notifications_android_tv import ConnectError, InvalidResponse, Notifications


@pytest.mark.asyncio
async def test_timeout(httpx_mock: HTTPXMock) -> None:
    """Test if the connection is hitting the timeout."""

    def raise_timeout(request):
        """Set the timeout for the requests."""
        raise httpx.ReadTimeout(
            f"Unable to read within {request.extensions['timeout']}", request=request
        )

    httpx_mock.add_callback(raise_timeout)

    with pytest.raises(ConnectError):
        notifier = Notifications("0.0.0.0")
        await notifier.async_connect()


@pytest.mark.asyncio
async def test_sending_failed(httpx_mock: HTTPXMock) -> None:
    """Test sending a message fails."""
    httpx_mock.add_response(status_code=400)

    notifier = Notifications("0.0.0.0")
    with pytest.raises(InvalidResponse):
        await notifier.async_send("Message text")


@pytest.mark.asyncio
async def test_sending_successfull(httpx_mock: HTTPXMock) -> None:
    """Test sending a message is successful."""
    httpx_mock.add_response(status_code=200)

    notifier = Notifications("0.0.0.0")
    await notifier.async_send("Message text")



"""Exceptions raised by the library."""


class NotificationException(Exception):
    """Base class for all exceptions raised by the library."""


class ConnectError(NotificationException):
    """Exception raised for connection error."""


class InvalidResponse(NotificationException):
    """Exception raised for invalid response."""


class InvalidImage(NotificationException):
    """Exception raised for invalid image."""


class InvalidImageData(NotificationException):
    """Exception raised for invalid image data is provided."""



"""Example scripts for sending notifications."""

import asyncio
import logging

from notifications_android_tv import ConnectError, Notifications
from notifications_android_tv.exceptions import NotificationException
from notifications_android_tv.notifier import NotificationParams

_LOGGER = logging.getLogger(__name__)

HOST = "<host>"
ICON = "<icon path or url>"
IMAGE = "<image path or url>"


async def main() -> None:
    """Run the example script."""
    notifier = Notifications(HOST)

    # validate connection
    try:
        await notifier.async_connect()
    except ConnectError as err:
        _LOGGER.error(err)
        return

    # # Send a basic notification with message only
    await notifier.async_send("This is a simple notification message")

    # For constructing paramters from string values as documented
    # in Home Assistant https://www.home-assistant.io/integrations/nfandroidtv
    try:
        notification_params = NotificationParams.from_dict(
            {
                "duration": "10",
                "color": "red",
                "fontsize": "small",
                "position": "bottom-right",
                "transparency": "25%",
                "interrupt": 0,
                "icon": {"path": ICON},
                "image": {"url": IMAGE},
            }
        )
    except ValueError as err:
        _LOGGER.error(err)
        return
    try:
        await notifier.async_send(
            "This is a notification message",
            title="Notification Title",
            params=notification_params,
        )
    except NotificationException as err:
        _LOGGER.error(err)


if __name__ == "__main__":
    asyncio.run(main())



"""Notification class for Android TV."""

import base64
import logging
from typing import Any

import httpx

from .const import (
    DEFAULT_ICON,
    DEFAULT_TITLE,
)
from .exceptions import ConnectError, InvalidResponse
from .helpers import NotificationParams

_LOGGER = logging.getLogger(__name__)


class Notifications:
    """Notifications class for Android/Fire Tvs."""

    def __init__(
        self,
        host: str,
        port: int = 7676,
        httpx_client: httpx.AsyncClient | None = None,
    ) -> None:
        """Initialize notifier."""
        self.url = f"http://{host}:{port}"
        self.httpx_client = httpx_client

    async def async_connect(self) -> None:
        """Test connecting to server."""
        httpx_client: httpx.AsyncClient = self.httpx_client or httpx.AsyncClient(
            verify=False
        )
        try:
            async with httpx_client as client:
                await client.get(self.url, timeout=30)
        except (httpx.ConnectError, httpx.TimeoutException) as err:
            raise ConnectError(f"Connection to {self.url} failed: {err}") from err

    async def async_send(
        self,
        message: str,
        *,
        title: str | None = None,
        params: NotificationParams | None = None,
    ) -> None:
        """Send message with parameters.

        :param message: The notification message.
        :param title: (Optional) The notification title.
        :params params: (Optional) Notification parameters. Construct using NotificationParams.

        Usage:
        >>> from notifications_android_tv import Notifications
        >>> notifier = Notifications("192.168.3.88")
        >>> await notifier.async_connect()
        >>> await notifier.async_send(
                "message to be sent",
                title="Notification title",
                params=NotificationParams(
                    duration=5,
                    position=Positions.BOTTOM_RIGHT,
                    fontsize=FontSize.MEDIUM,
                )
            )
        """
        data: dict[str, Any] = {
            "msg": message,
            "title": title or DEFAULT_TITLE,
        }
        if params is not None:
            data.update(params.data_params)
            files = await params.get_images()
        else:
            icon_bytes = base64.b64decode(DEFAULT_ICON)
            files = {
                "filename": (
                    "image",
                    icon_bytes,
                    "application/octet-stream",
                    {"Expires": "0"},
                )
            }

        _LOGGER.debug("data: %s, files: %s", data, files)

        httpx_client: httpx.AsyncClient = self.httpx_client or httpx.AsyncClient(
            verify=False
        )
        try:
            async with httpx_client as client:
                response = await client.post(
                    self.url, data=data, files=files, timeout=10
                )

        except (httpx.ConnectError, httpx.TimeoutException) as err:
            raise ConnectError(f"Error communicating with {self.url}: {err}") from err
        if response.status_code != httpx.codes.OK:
            raise InvalidResponse(f"Error sending message: {response}")
